/**************************************************************************
 * clients.c
 *
 * Client functions
 *
 * Modified: 2008.076
 *
 * Chad Trabant, IRIS Data Management Center
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <pthread.h>

#include <libdali.h>

#include "dali2liss.h"
#include "clients.h"
#include "logging.h"

/* Define the number of no-action loops that trigger the throttle */
#define THROTTLE_TRIGGER 10

static int RecvData (ClientInfo *cinfo, char *buffer, size_t buflen);


/***********************************************************************
 * ClientThread:
 *
 * Thread to handle all communications with a SeedLink client.
 *
 * Returns NULL.
 ***********************************************************************/
void *
ClientThread (void *arg)
{
  ClientInfo *cinfo;
  struct thread_data *mytdp;
  int sockflags;
  int setuperr = 0;
  int nread;
  int rv;

  /* Client thread specific packet header and data buffers */
  DLPacket packet;
  char packetdata[MAXPACKETSIZE];
  char buffer[10];
  
  /* Throttle related */
  char throttle = 0; /* Controls throttling of main loop */
  fd_set readset;    /* File descriptor set for select() */
  struct timeval timeout;  /* Timeout throttle for select() */
  
  mytdp = (struct thread_data *) arg;
  cinfo = (ClientInfo *) mytdp->td_prvtptr;
  
  /* Connect to DataLink server if needed */
  if ( cinfo->dlcp->link == -1 )
    {
      /* Connect to server */
      if ( dl_connect (cinfo->dlcp) < 0 )
	{
	  fprintf (stderr, "[%s] Error connecting to DataLink server\n",
		   cinfo->addrstr);
	  setuperr = 1;
	}
      
      /* Send match pattern if supplied */
      if ( cinfo->match && ! setuperr )
	{
	  if ( dl_match (cinfo->dlcp, cinfo->match) < 0 )
	    setuperr = 1;
	}
      
      /* Send reject pattern if supplied */
      if ( cinfo->reject && ! setuperr )
	{
	  if ( dl_reject (cinfo->dlcp, cinfo->reject) < 0 )
	    setuperr = 1;
	}
      
    }
  
  /* Set client socket connection to non-blocking */
  sockflags = fcntl(cinfo->socket, F_GETFL, 0);
  sockflags |= O_NONBLOCK;
  if ( fcntl(cinfo->socket, F_SETFL, sockflags) == -1 )
    {
      lprintf (0, "[%s] Error setting non-blocking flag: %s",
	       cinfo->addrstr, strerror(errno));
      setuperr = 1;
    }
  
  /* Shutdown the client connection if there were setup errors */
  if ( setuperr )
    {
      /* Close client socket */
      if ( cinfo->socket )
        {
          close (cinfo->socket);
          cinfo->socket = 0;
        }
      
      /* Set thread closing status */
      pthread_mutex_lock (&(mytdp->td_lock));
      mytdp->td_flags = TDF_CLOSING;
      pthread_mutex_unlock (&(mytdp->td_lock));
      
      return NULL;
    }
  
  /* Set thread active status */
  pthread_mutex_lock (&(mytdp->td_lock));
  if ( mytdp->td_flags == TDF_SPAWNING )
    mytdp->td_flags = TDF_ACTIVE;
  pthread_mutex_unlock (&(mytdp->td_lock));
  
  /* Main client loop, delegating requests and handling data flow */
  while ( mytdp->td_flags != TDF_CLOSE )
    {
      /* Increment throttle trigger count by default */
      if ( throttle < THROTTLE_TRIGGER )
	throttle++;
      
      /* Check for data from client */
      nread = RecvData (cinfo, buffer, sizeof(buffer));
      
      /* Success or error, either way we are done */
      if ( nread )
	{
	  if ( nread > 0 )
	    lprintf (0, "[%s] Data received from client, exiting", cinfo->addrstr);
	  
	  break;
	}
      
      /* Check for packet from DataLink server */
      rv = dl_collect_nb (cinfo->dlcp, &packet, packetdata, sizeof(packetdata), 0);
      
      /* Forward any recieved packet to the client */
      if ( rv == DLPACKET )
	{
	  char OKGO = 1;
	  
	  lprintf (3, "[%s] Recevied %s (%u bytes) from DataLink server",
		   cinfo->addrstr, packet.streamid, packet.datasize);
	  
	  /* If data do not throttle */
	  throttle = 0;
	  
	  if ( dl_splitstreamid (packet.streamid, NULL, NULL, NULL, NULL, buffer) )
	    {
	      lprintf (0, "[%s] Error parsing packet type from %s",
		       cinfo->addrstr, buffer);
	      OKGO = 0;
	    }
	  
	  if ( OKGO && (strncmp ("MSEED", buffer, sizeof(buffer)) ||
			packet.datasize != 512) )
	    {
	      lprintf (0, "[%s] Skipping packet type %s (%d bytes)",
		       cinfo->addrstr, buffer, packet.datasize);
	    }
	  else if ( OKGO )
	    {
	      /* Send Mini-SEED record to client */
	      if ( send (cinfo->socket, packetdata, packet.datasize, 0) < 0 )
		{
		  lprintf (0, "[%s] Error sending Mini-SEED record to client",
			   cinfo->addrstr);
		  break;
		}
	    }
	}
      /* If there was no next packet, turn on immediate throttling */
      else if ( rv == DLNOPACKET )
	{
	  throttle = THROTTLE_TRIGGER;
	}
      /* Otherwise the connection was shutdown or an error occurred */
      else
	{
	  break;
	}
      
      /* Throttle loop if THROTTLE_TRIGGER or more loops without action */
      if ( throttle >= THROTTLE_TRIGGER )
	{
	  /* Throttle the loop using select() so that data from the
	     client is never waiting for a timeout */
	  
	  /* Configure the read descriptor set with only our server socket */
	  FD_ZERO(&readset);
	  FD_SET(cinfo->dlcp->link, &readset);
	  
	  /* Timeout (throttle when no data) is 1/10 of a second */
	  timeout.tv_sec = 0;
	  timeout.tv_usec = 100000;
	  
	  select (cinfo->dlcp->link+1, &readset, NULL, NULL, &timeout);
	}
    } /* End of main client loop */
  
  /* Set thread CLOSING status, locking entire client list */
  pthread_mutex_lock (&cthreads_lock);
  mytdp->td_flags = TDF_CLOSING;
  pthread_mutex_unlock (&cthreads_lock);
  
  /* Close connection to DataLink server */
  if ( cinfo->dlcp->link != -1 )
    {
      dl_disconnect (cinfo->dlcp);
    }
  
  /* Close client socket */
  if ( cinfo->socket )
    {
      close (cinfo->socket);
      cinfo->socket = 0;
    }
  
  lprintf (1, "Client disconnected: %s", cinfo->addrstr);
  
  /* Set thread CLOSED status */
  pthread_mutex_lock (&(mytdp->td_lock));
  mytdp->td_flags = TDF_CLOSED;
  pthread_mutex_unlock (&(mytdp->td_lock));
  
  return NULL;
}  /* End of ClientThread() */


/***********************************************************************
 * RecvData:
 *
 * Read up to buflen characters from a socket and place them into
 * buffer.
 *
 * The buffer provided must already be allocated and have enough space
 * for buflen bytes.
 *
 * Return number of characters read on success, 0 when no data for
 * non-blocking socket, -1 on connection shutdown and -2 on error.
 ***********************************************************************/
int
RecvData (ClientInfo *cinfo, char *buffer, size_t buflen)
{
  int nrecv;
  int nread = 0;
  char *bptr = buffer;
  
  /* Recv until buflen bytes have been read */
  while ( nread < buflen )
    {
      if ( (nrecv = recv(cinfo->socket, bptr, buflen-nread, 0)) < 0 )
        {
	  /* The only acceptable error is EAGAIN (no data on non-blocking) */
	  if ( errno == EAGAIN )
	    {
	      return 0;
	    }
	  
	  return -2;
	}
      
      /* Peer completed an orderly shutdown */
      if ( nrecv == 0 )
	return -1;
      
      /* Update recv pointer and byte count */
      if ( nrecv > 0 )
	{
	  bptr += nrecv;
	  nread += nrecv;
	}
    }
  
  return nread;
}  /* End of RecvData() */
