/**************************************************************************
 * dali2liss.c
 *
 * Multi-threaded TCP bridge between DataLink and LISS.
 *
 * Modified: 2010.010
 *
 * Chad Trabant, IRIS Data Management Center
 **************************************************************************/

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <time.h>
#include <unistd.h>

#include <libdali.h>

#include "logging.h"
#include "clients.h"
#include "dali2liss.h"

#define PACKAGE   "dali2liss"
#define VERSION   "2010.110"

/* A structure to list IP addresses ranges */
typedef struct IPNet_s
{
  struct in_addr network;
  struct in_addr netmask;
  struct IPNet_s *next;
} IPNet;

/* Global variables defined here */
pthread_mutex_t cthreads_lock = PTHREAD_MUTEX_INITIALIZER;
struct cthread *cthreads = 0;      /* Client threads list */
int      shutdownsig = 0;          /* Shutdown signal */

/* Local functions and variables */
static struct thread_data *InitThreadData (void *prvtptr);
static void    *ListenThread (void *arg);
static int      InitServerSocket (unsigned short port);
static int      ProcessParam (int argcount, char **argvec);
static char    *GetOptVal (int argcount, char **argvec, int argopt);
static int      AddIPNet (IPNet **pplist, char *network);
static int      MatchIP (IPNet *list, struct in_addr *addr);
static int      IsAllDigits (char *string);
static void     TermHandler();
static void     PrintHandler();
static void     Usage ();

static unsigned short serverport = 4000; /* LISS server port to listen on */
static int      clientcount   = 0;       /* Track number of connected clients */
static uint32_t maxclients    = 100;     /* Enforce maximum number of clients */
static char    *dladdress     = 0;       /* Address of DataLink server */
static int      dlkeepalive   = -1;      /* Keepalive interval for DataLink */
static char    *matchpattern  = 0;       /* Source ID matching expression */
static char    *rejectpattern = 0;       /* Source ID rejecting expression */

static IPNet *rejectips = 0;

int
main (int argc, char* argv[])
{
  int serversocket = 0;
  time_t curtime;
  time_t chktime;
  struct timespec timereq;
  struct cthread *ctp;
  struct cthread *loopctp;
  char statusstr[100];
    
  pthread_t listenThread = 0;
  struct thread_data *listenThreadDP = 0;
  
  /* Signal handling using POSIX routines */
  struct sigaction sa;
  
  sa.sa_flags = SA_RESTART;
  sigemptyset(&sa.sa_mask);
  
  sa.sa_handler = PrintHandler;
  sigaction(SIGUSR1, &sa, NULL);
  
  sa.sa_handler = TermHandler;
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  
  sa.sa_handler = SIG_IGN;
  sigaction(SIGHUP, &sa, NULL);
  sigaction(SIGPIPE, &sa, NULL);
  
  /* Process command line parameters */
  if (ProcessParam (argc, argv) < 0)
    return 1;
  
  /* Init the LISS server socket on the specified port */
  if ( (serversocket = InitServerSocket (serverport)) == -1 )
    {
      return 1;
    }
  
  /* Initialize watchdog loop interval timers */
  curtime = time (NULL);
  chktime = curtime;
  
  /* Set loop interval check tick to 1/4 second */
  timereq.tv_sec = 0;
  timereq.tv_nsec = 250000000;
  
  /* Watchdog loop: monitors the listenThread and client
     threads doing restarts and cleanup when necessary. */
  for (;;)
    {
      /* If shutdown is requested signal all client threads */
      if ( shutdownsig == 1 )
	{
	  shutdownsig = 2;
	  
	  /* Set shutdown loop throttle of .1 seconds */
	  timereq.tv_nsec = 100000000; 
	  
	  /* Close server socket, causing the listen thread to exit too */
	  if ( serversocket > 0 )
	    {
	      lprintf (3, "Closing server socket");
	      shutdown (serversocket, SHUT_RDWR);
	      close (serversocket);
	      serversocket = -1;
	    }
	  
	  /* Request shutdown of client threads */
	  pthread_mutex_lock (&cthreads_lock);
	  loopctp = cthreads;
	  while ( loopctp )
	    {
	      if ( ! (loopctp->td->td_flags & TDF_CLOSING) && ! (loopctp->td->td_flags & TDF_CLOSED) )
		{
		  lprintf (3, "Requesting shutdown of client thread %lu",
			   (unsigned long int)loopctp->td->td_id);
		  
		  pthread_mutex_lock (&(loopctp->td->td_lock));
		  loopctp->td->td_flags = TDF_CLOSE;
		  pthread_mutex_unlock (&(loopctp->td->td_lock));
		}
	      loopctp = loopctp->next;
	    }
	  pthread_mutex_unlock (&cthreads_lock);
	}
      
      if ( shutdownsig > 1 )
	{
	  /* Safety valve for deadlock, should never get here */
	  if ( shutdownsig >= 50 )
	    {
	      lprintf (0, "Shutdown did not complete cleanly after ~5 seconds");
	      break;
	    }
	  
	  shutdownsig++;
	}
      
      /* Report status of listen thread */
      if ( listenThread && listenThreadDP )
	{
	  sprintf (statusstr, "listenThread %lu flags:", (unsigned long int) listenThread);
	  if ( listenThreadDP->td_flags & TDF_SPAWNING ) strcat (statusstr, " SPAWNING");
	  if ( listenThreadDP->td_flags & TDF_ACTIVE ) strcat (statusstr, " ACTIVE");
	  if ( listenThreadDP->td_flags & TDF_CLOSE ) strcat (statusstr, " CLOSE");
	  if ( listenThreadDP->td_flags & TDF_CLOSING ) strcat (statusstr, " CLOSING");
	  if ( listenThreadDP->td_flags & TDF_CLOSED ) strcat (statusstr, " CLOSED");
	  lprintf (3, "%s", statusstr);
	}
      else
	lprintf (2, "Listening thread not running");
      
      /* Cleanup listen thread if it is CLOSED */
      if ( listenThread && listenThreadDP )
	{
	  if ( listenThreadDP->td_flags & TDF_CLOSED )
	    {
	      lprintf (1, "Joining CLOSED listen thread");
	      
	      if ( (errno=pthread_join (listenThreadDP->td_id, NULL)) )
		{
		  lprintf (0, "Error joining CLOSED listen thread %lu: %s",
			   (unsigned long int)listenThreadDP->td_id, strerror(errno));
		}
	      
	      free (listenThreadDP);
	      listenThread = 0;
	      listenThreadDP = 0;
	    }
	}
      
      /* Start new listening thread if needed */
      if ( listenThread == 0 && ! shutdownsig )
	{
	  if ( ! (listenThreadDP = InitThreadData(&serversocket)) )
	    {
	      lprintf (0, "Error initializing thread_data: %s", strerror(errno));
	    }
	  else
	    {
	      lprintf (2, "Starting listen thread");
	      
	      if ( (errno=pthread_create (&listenThread, NULL, ListenThread, (void*) listenThreadDP)) )
		{
		  lprintf (0, "Error creating listen thread: %s", strerror(errno));
		  listenThread = 0;
		  if ( listenThreadDP ) free (listenThreadDP);
		  listenThreadDP = 0;
		}
	      else
		listenThreadDP->td_id = listenThread;
	    }
	}
      
      /* Loop through client thread list printing status and doing cleanup */
      pthread_mutex_lock (&cthreads_lock);
      loopctp = cthreads;
      while ( loopctp ) 
	{
	  ctp = loopctp;
	  loopctp = loopctp->next;
	  
	  sprintf (statusstr, "Client Thread %lu Flags:", (unsigned long int)ctp->td->td_id);
	  if ( ctp->td->td_flags & TDF_SPAWNING ) strcat (statusstr, " SPAWNING");
	  if ( ctp->td->td_flags & TDF_ACTIVE ) strcat (statusstr," ACTIVE");
	  if ( ctp->td->td_flags & TDF_CLOSING ) strcat (statusstr," CLOSING");
	  if ( ctp->td->td_flags & TDF_CLOSED ) strcat (statusstr," CLOSED");
	  lprintf (3, "%s", statusstr);
	  
	  /* Free associated resources and join CLOSED client threads */
	  if ( ctp->td->td_flags & TDF_CLOSED )
	    {
	      lprintf (3, "Removing client thread %lu from the cthreads list",
		       (unsigned long int)ctp->td->td_id);
	      
	      /* Unlink from the cthreads list */
	      if ( ! ctp->prev && ! ctp->next ) cthreads = 0;
	      if ( ! ctp->prev && ctp->next ) cthreads = ctp->next;
	      if ( ctp->prev ) ctp->prev->next = ctp->next;
	      if ( ctp->next ) ctp->next->prev = ctp->prev;
	      
	      if ( (errno=pthread_join(ctp->td->td_id, NULL)) )
		{
		  lprintf (0, "Error joining CLOSED thread %lu: %s",
			   (unsigned long int)ctp->td->td_id, strerror(errno));
		}
	      
	      /* Free the ClientInfo structure stored at the prvtptr */
	      if ( ctp->td->td_prvtptr )
		free (ctp->td->td_prvtptr);
	      
	      /* Free thread data structure */
	      if ( ctp->td )
		free (ctp->td);
	      
	      /* Decrement client count */
	      if ( clientcount > 0 )
		clientcount--;
	     
	      /* Free thread data */
	      free (ctp);
	    }
	}
      pthread_mutex_unlock (&cthreads_lock);
      
      lprintf (3, "Client connections: %d", clientcount);
      
      /* All done if shutting down and no threads left */
      if ( shutdownsig >= 2 && clientcount == 0 && ! listenThread )
	break;
      
      /* Throttle the loop during shutdown */
      if ( shutdownsig )
	{
	  nanosleep (&timereq, NULL);
	}
      /* Otherwise, throttle the loop for a second, signals will interrupt */
      else
	{
	  while ( ((curtime = time (NULL)) - chktime) < 1 && ! shutdownsig )
	    nanosleep (&timereq, NULL);
	}
      
      chktime = curtime;	      
    }  /* End of main watchdog loop */
  
  /* Close listening sockets if not already closed */
  if ( serversocket > 0 )
    {
      lprintf (3, "Closing server socket");
      close (serversocket);
    }
  
  return 0;
}  /* End of main() */


/***********************************************************************
 * InitThreadData:
 *
 * Initialize thread data.
 *
 * Return pointer to thread_data on success and 0 on error.
 ***********************************************************************/
struct thread_data *
InitThreadData (void *prvtptr)
{
  struct thread_data *rtdp;
  
  rtdp = (struct thread_data *) malloc (sizeof(struct thread_data));
  
  if ( ! rtdp )
    {
      lprintf (0, "Error malloc'ing thread_data: %s", strerror(errno));
      return 0;
    }
  
  if ( pthread_mutex_init(&(rtdp->td_lock), NULL) )
    {
      lprintf (0, "Error initializing thread_data mutex: %s", strerror(errno));
      return 0;
    }
  
  rtdp->td_id = 0;
  rtdp->td_flags = TDF_SPAWNING;
  
  rtdp->td_prvtptr = prvtptr;
  
  return rtdp;
}  /* End of InitThreadData() */


/***********************************************************************
 * ListenThread:
 *
 * Thread to accept connections and dispatch client threads.
 *
 * Return NULL.
 ***********************************************************************/
void *
ListenThread (void *arg)
{
  pthread_t ctid;
  unsigned short serverport;
  int serversocket;
  int clientsocket;
  struct thread_data *mytdp;
  struct thread_data *tdp;
  struct cthread *ctp;
  ClientInfo *cinfo;
  DLCP *dlconn;
  
  char subid[100];
  
  struct sockaddr_in inet_addr;
  socklen_t addr_len = sizeof(inet_addr);
  
  mytdp = (struct thread_data *) arg;
  serversocket = *((int *)mytdp->td_prvtptr);
  
  /* Set thread active status */
  pthread_mutex_lock (&(mytdp->td_lock));
  mytdp->td_flags = TDF_ACTIVE;
  pthread_mutex_unlock (&(mytdp->td_lock));
  
  /* Lookup port we are listening on */
  getsockname (serversocket, (struct sockaddr *)&inet_addr, &addr_len);
  serverport = ntohs (inet_addr.sin_port);
  
  lprintf (1, "Listening for connections on port %d", serverport);
  
  /* Enter connection dispatch loop, spawning a new thread for each incoming connection */
  while ( ! shutdownsig )
    {
      /* Process next connection in queue */
      clientsocket = accept (serversocket, (struct sockaddr *)&inet_addr, &addr_len);
      
      /* Check for accept errors */
      if ( clientsocket == -1 )
	{
	  /* If not shutting down this is a connection error */
	  if ( ! shutdownsig )
	    lprintf (0, "Could not accept incoming connection: %s", strerror(errno));
	  
	  break;
	}
      
      /* Reject clients in the rejection list */
      if ( rejectips )
	{
	  if ( MatchIP (rejectips, &(inet_addr.sin_addr)) )
	    {
	      lprintf (1, "Rejecting connection from: %d.%d.%d.%d:%d",
		       NIPQUAD(inet_addr.sin_addr.s_addr), ntohs(inet_addr.sin_port));
	      close (clientsocket);
	      continue;
	    }
	}
      
      /* Enforce maximum number of clients if specified */
      if ( maxclients && clientcount >= maxclients )
	{
	  lprintf (1, "Maximum number of clients exceeded: %d", maxclients);
	  lprintf (1, "  Rejecting connection from: %d.%d.%d.%d:%d",
		   NIPQUAD(inet_addr.sin_addr.s_addr), ntohs(inet_addr.sin_port));
	  close (clientsocket);
	  continue;
	}
      
      /* Generate sub-client ID and use a program name */
      snprintf (subid, sizeof(subid), "LISS:%d-%d.%d.%d.%d:%d|%s", serverport,
		NIPQUAD(inet_addr.sin_addr.s_addr), ntohs(inet_addr.sin_port), PACKAGE);
      
      /* Allocate and initialize DataLink connection parameters */
      if ( ! (dlconn = dl_newdlcp (dladdress, subid)) )
	{
	  lprintf (0, "Error initializing DataLink connection parameters");
	  break;
	}
      
      /* Set DataLink keepalive interval, allowing for valid 0 setting */
      if ( dlkeepalive >= 0 )
	dlconn->keepalive = dlkeepalive;
      
      /* Allocate and initialize connection info struct */
      if ( (cinfo = (ClientInfo *) malloc (sizeof(ClientInfo))) == NULL )
	{
	  lprintf (0, "Error allocating memory for connection info");
	  break;
	}
      
      /* Initialize ClientInfo to zeros */
      memset (cinfo, 0, sizeof(ClientInfo));
      
      cinfo->socket = clientsocket;
      cinfo->dlcp = dlconn;
      
      cinfo->match = matchpattern;
      cinfo->reject = rejectpattern;
      
      /* Create address:port string */
      snprintf (cinfo->addrstr, sizeof(cinfo->addrstr), "%d.%d.%d.%d:%d",
		NIPQUAD(inet_addr.sin_addr.s_addr), ntohs(inet_addr.sin_port));
      
      lprintf (1, "Client connected: %s", cinfo->addrstr);
      
      /* Create new client thread */
      if ( ! (tdp = InitThreadData (cinfo)) )
	{
	  lprintf (0, "Error initializing thread_data: %s", strerror(errno));
	  if ( clientsocket )
	    close (clientsocket);
	  return NULL;
	}
      
      if ( (errno=pthread_create (&ctid, NULL, ClientThread, (void*) tdp)) )
	{
	  lprintf (0, "Error creating new client thread: %s", strerror(errno));
	  if ( clientsocket )
	    close (clientsocket);
	  if ( tdp )
	    free (tdp);
	  tdp = 0;
	}
      else
	{
	  /* Update thread id, no locking, should be safe */
	  tdp->td_id = ctid;
	  
	  ctp = (struct cthread *) malloc (sizeof (struct cthread));
	  if ( ctp == NULL ) {
	    lprintf (0, "Error malloc'ing cthread: %s", strerror(errno));
	    return NULL;
	  }
	  
	  ctp->td = tdp;
	  ctp->prev = 0;
	  
	  /* Add ctp to the beginning of the client threads list (cthreads) */
	  pthread_mutex_lock (&cthreads_lock);
	  if ( cthreads ) { ctp->next = cthreads; cthreads->prev = ctp; }
	  else { ctp->next = 0; }
	  cthreads = ctp;
	  pthread_mutex_unlock (&cthreads_lock);
	  
	  /* Increment client count */
	  clientcount++;
	}
    }
  
  /* Set thread closing status */
  pthread_mutex_lock (&(mytdp->td_lock));
  mytdp->td_flags = TDF_CLOSED;
  pthread_mutex_unlock (&(mytdp->td_lock));
  
  lprintf (1, "Listening thread closing");
  
  return NULL;
}  /* End of ListenThread() */


/***********************************************************************
 * InitServerSocket:
 *
 * Initialize a TCP server socket on the specified port bound to all
 * local addresses/interfaces.
 *
 * Return socket descriptor on success and -1 on error.
 ***********************************************************************/
static int
InitServerSocket (unsigned short port)
{
  struct sockaddr_in addr;
  int fd;
  int optval;
  
  /* Create a socket */
  fd = socket (AF_INET, SOCK_STREAM, 0);
  if ( fd == -1 )
    {
      lprintf (0, "Error with socket(): %s", strerror(errno));
      return -1;
    }
  
  optval = 1;
  if ( setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) )
    {
      lprintf (0, "Error with setsockopt(): %s", strerror(errno));
      return -1;
    }
  
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(port);
  
  if ( bind (fd, (struct sockaddr*) &addr, sizeof addr) == -1 )
    {
      lprintf (0, "Error with bind(): %s", strerror(errno));
      close (fd);
      return -1;
    }

  if ( listen (fd, 5) == -1 )
    {
      lprintf (0, "Error with listen(): %s", strerror(errno));
      close (fd);
      return -1;
    }

  return fd;
}  /* End of InitServerSocket() */


/***************************************************************************
 * ProcessParam:
 *
 * Process the command line parameters.
 *
 * Returns 0 on success, and -1 on failure
 ***************************************************************************/
static int
ProcessParam (int argcount, char **argvec)
{
  char *rejectip;
  int optind;
  
  /* Process all command line arguments */
  for (optind = 1; optind < argcount; optind++)
    {
      if (strcmp (argvec[optind], "-V") == 0)
        {
          fprintf (stderr, "%s version: %s\n", PACKAGE, VERSION);
          exit (0);
        }
      else if (strcmp (argvec[optind], "-h") == 0)
        {
          Usage();
        }
      else if (strncmp (argvec[optind], "-v", 2) == 0)
        {
          verbose += strspn (&argvec[optind][1], "v");
        }
      else if (strcmp (argvec[optind], "-k") == 0)
        {
          dlkeepalive = strtoul (GetOptVal(argcount, argvec, optind++), NULL, 10);
        }
      else if (strcmp (argvec[optind], "-m") == 0)
        {
          matchpattern = GetOptVal(argcount, argvec, optind++);
        }
      else if (strcmp (argvec[optind], "-r") == 0)
        {
          rejectpattern = GetOptVal(argcount, argvec, optind++);
        }
      else if (strcmp (argvec[optind], "-C") == 0)
        {
	  maxclients = strtoul (GetOptVal(argcount, argvec, optind++), NULL, 10);
        }
      else if (strcmp (argvec[optind], "-p") == 0)
        {
	  serverport = strtoul (GetOptVal(argcount, argvec, optind++), NULL, 10);
        }
      else if (strcmp (argvec[optind], "-R") == 0)
        {
	  /* Add IP and netmask to rejection IP list */
	  rejectip = GetOptVal(argcount, argvec, optind++);
	  if ( AddIPNet (&rejectips, rejectip) )
	    {
	      lprintf (0, "Error adding %s to rejection list", rejectip);
	      return -1;
	    }
        }
      else if (strncmp (argvec[optind], "-", 1) == 0)
        {
          lprintf (0, "Unknown option: %s", argvec[optind]);
          exit (1);
        }
      else if ( ! dladdress )
        {
          dladdress = argvec[optind];
        }
      else
	{
          fprintf(stderr, "Unknown option: %s\n", argvec[optind]);
          exit (1);
	}
    }
  
  /* Make sure a server was specified */
  if ( ! dladdress )
    {
      fprintf(stderr, "No DataLink server specified\n\n");
      fprintf(stderr, "%s version: %s\n\n", PACKAGE, VERSION);
      fprintf(stderr, "Usage: %s [options] [host][:][port]\n", PACKAGE);
      fprintf(stderr, "Try '-h' for detailed help\n");
      exit (1);
    }
  
  /* Initialize the libdali logging facitility, use lprint() bridge to lprintf() */
  dl_loginit (verbose, lprint, NULL, lprint, NULL);
  
  /* Report the program version */
  lprintf (0, "%s version: %s", PACKAGE, VERSION);
  
  /* Load the match stream list from a file if the argument starts with '@' */
  if ( matchpattern && *matchpattern == '@' )
    {
      char *filename = matchpattern + 1;
      
      if ( ! (matchpattern = dl_read_streamlist (NULL, filename)) )
        {
          lprintf (2, "Cannot read matching list file: %s", filename);
          exit (1);
        }
    }
  
  /* Load the reject stream list from a file if the argument starts with '@' */
  if ( rejectpattern && *rejectpattern == '@' )
    {
      char *filename = rejectpattern + 1;
      
      if ( ! (rejectpattern = dl_read_streamlist (NULL, filename)) )
        {
          lprintf (2, "Cannot read rejecting list file: %s", filename);
          exit (1);
        }
    }
  
  return 0;
}  /* End of ProcessParam() */


/***************************************************************************
 * GetOptVal:
 * Return the value to a command line option; checking that the value is
 * itself not an option (starting with '-') and is not past the end of
 * the argument list.
 *
 * argcount: total arguments in argvec
 * argvec: argument list
 * argopt: index of option to process, value is expected to be at argopt+1
 *
 * Returns value on success and exits with error message on failure
 ***************************************************************************/
static char *
GetOptVal (int argcount, char **argvec, int argopt)
{
  if ( argvec == NULL || argvec[argopt] == NULL )
    {
      lprintf (0, "GetOptVal(): NULL option requested");
      exit (1);
    }

  if ( (argopt+1) < argcount && *argvec[argopt+1] != '-' )
    return argvec[argopt+1];

  lprintf (0, "Option %s requires a value", argvec[argopt]);
  exit (1);
}  /* End of GetOptVal() */


/***************************************************************************
 * AddIPNet:
 *
 * Add an IPV4 network and netmask to an IPNet list.
 * 
 * Returns 0 on success and -1 on error.
 ***************************************************************************/
static int
AddIPNet (IPNet **pplist, char *network)
{
  IPNet *newipnet;
  char net[100];
  char *mask;
  
  if ( ! pplist || ! network )
    return -1;
  
  /* Allocate new IPNet */
  if ( ! (newipnet = (IPNet *) malloc (sizeof(IPNet))) )
    {
      lprintf (0, "AddIPNet(): Error allocation memory for IPNet");
      return -1;
    }
  
  /* Copy network string for manipulation */
  strncpy (net, network, sizeof(net));
  
  /* Split netmask from network: "IP/netmask" */
  if ( (mask = strchr (net, '/')) )
    {
      *mask++ = '\0';
    }
  else
    {
      /* The default netmask */
      mask = "32";
    }

  /* Convert bit count netmask to binary format */
  if ( IsAllDigits (mask) )
    {
      uint32_t bitsmask = 0;
      unsigned long int numbits = strtoul (mask, NULL, 10);
      int i;

      /* Sanity check and catch errors from strtoul() */
      if ( numbits > 32 )
	{
	  lprintf (0, "bit-count netmask (%s) must be less than 32", mask);
	  return -1;
	}
      
      for (i = 0; i < numbits; i++)
        {
	  bitsmask = bitsmask >> 1;
	  bitsmask |= 0x80000000;
        }
      
      bitsmask = htonl (bitsmask);
      
      /* Assign bit count-derived mask to the netmask address */
      newipnet->netmask.s_addr = bitsmask;
    }
  else
    {
      /* Convert netmask string to in_addr, supporting only AF_INET */
      if ( inet_pton (AF_INET, mask, &(newipnet->netmask)) <= 0 )
	{
	  lprintf (0, "Error parsing net mask address: %s", mask);
	  return -1;
	}
    }
  
  /* Convert IP string to in_addr, supporting only AF_INET */
  if ( inet_pton (AF_INET, net, &(newipnet->network)) <= 0 )
    {
      lprintf (0, "Error parsing IP address: %s", net);
      return -1;
    }
  
  /* AND the ip and netmask to get the actual network */
  newipnet->network.s_addr &= newipnet->netmask.s_addr;
  
  /* Push the new entry on the top of the list */
  newipnet->next = *pplist;
  *pplist = newipnet;
  
  return 0;
}  /* End of AddIPNet() */


/***************************************************************************
 * MatchIP:
 *
 * Search the specified IPNet list for an entry that matches the given
 * IP address.
 * 
 * Returns 1 if match found and 0 if no match found.
 ***************************************************************************/
static int
MatchIP (IPNet *list, struct in_addr *addr)
{
  IPNet *net = list;
  struct in_addr testnet;
  
  if ( ! list )
    return 0;
  
  /* Search IPNet list for a matching entry for addr */
  while ( net )
    {
      /* Copy the addr to the test addr */
      testnet.s_addr = addr->s_addr;
      
      /* AND the test addr with the netmask to get the test network */
      testnet.s_addr &= net->netmask.s_addr;
      
      /* Check for match between test network and list network */
      if ( testnet.s_addr == net->network.s_addr )
	return 1;
      
      net = net->next;
    }
  
  return 0;
}  /* End of MatchIP() */


/*********************************************************************
 * IsAllDigits:
 *
 * Return 1 if the specified string is all digits and 0 otherwise.
 *********************************************************************/
static int
IsAllDigits (char *string)
{
  int idx;
  int length;

  length = strlen (string);
  
  for ( idx=0; idx < length; idx++ )
    {
      if ( ! strchr("0123456789", string[idx]) )
	return 0;
    }
  
  return 1;
}  /* End of IsAllDigits() */


/***************************************************************************
 * TermHandler and PrintHandler (USR1 signal):
 * Signal handler routines.
 ***************************************************************************/
static void
TermHandler (int sig)
{
  shutdownsig = 1;
}
static void
PrintHandler (int sig)
{
}

/***************************************************************************
 * Usage:
 *
 * Print usage message and exit.
 ***************************************************************************/
static void
Usage()
{
  fprintf(stderr,"%s version %s\n\n", PACKAGE, VERSION);
  fprintf(stderr,"Usage: %s [host][:][port]\n\n", PACKAGE);
  fprintf(stderr," ## Options ##\n"
          " -V             Report program version\n"
          " -h             Show this usage message\n"
          " -v             Be more verbose, multiple flags can be used\n"
	  " -k secs        Specify DataLink keepalive interval in seconds\n"
	  " -m match       DataLink stream match expression\n"
	  " -r reject      DataLink stream reject expression\n"
	  "                  Typical stream names: 'Net_Sta_Loc_Chan/MSEED'\n"
	  " -C maxclnt     Maximum number of concurrent clients (currently: %d)\n"
	  " -p port        Port to listen on for LISS connections (currently %hu)\n"
	  " -R ip/mask     Reject a specified IP or subnet\n"
	  "\n"
	  " [host][:][port]  Address of the DataLink server in host:port format\n"
	  "                  if host is omitted (i.e. ':16000'), localhost is assumed\n"
	  "                  if :port is omitted (i.e. 'localhost'), 16000 is assumed\n\n",
          maxclients, serverport);
  exit (1);
}  /* End of Usage() */
